Attribute VB_Name = "modDDEServer"
' ********************************************************************************                                                                             *
' *                                                                              *
' *  modDDEServer.bas - the core dde server                                      *
' *                                                                              *
' *  Copyright (c) 2000-2002 Asperon (asperon@home.se), last updated 2002-09-08  *
' *                                                                              *
' *  Based on the work by Eric D. Wilson(edwilson97@yahoo.com) and               *
' *  Peter Edwards.                                                              *
' *                                                                              *
' *  This program is free software; you can redistribute it and/or modify        *
' *  it under the terms of the GNU General Public License as published by        *
' *  the Free Software Foundation; either version 2 of the License, or           *
' *  any later version.                                                          *
' *                                                                              *
' *  This program is distributed in the hope that it will be useful,             *
' *  but WITHOUT ANY WARRANTY; without even the implied warranty of              *
' *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
' *  GNU General Public License for more details.                                *
' *                                                                              *
' *  You should have received a copy of the GNU General Public License           *
' *  along with this program; if not, write to the Free Software                 *
' *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   *
' *                                                                              *
' ********************************************************************************
Option Explicit

Private g_hszProgman As Long        ' String handle for the server name.
Private g_hszGroups As Long         ' String handle for the topic name.
Private g_hszFolders As Long        ' String handle for the Poke item name.
Private bRunning As Boolean         ' Server running flag.

' The DDE Instance ID and Item variables need to be global since there used by
' both the form and the code module.
Global lInstID As Long             ' DDE instance identifier.
Global g_hszAppProperties As Long    ' String handle for the item name.

'*************************************************************************
' DDEML Server Constants
'*************************************************************************
Public Const DDE_SERVER = "MyServer"
Public Const DDE_TOPIC = "MyTopic"
Public Const DDE_ADVISE = "MyAdvise"
Public Const DDE_REQUEST = "MyRequest"
Public Const DDE_POKE = "MyPoke"
Public Const DDE_COMMAND1 = "MAX"
Public Const DDE_COMMAND2 = "MIN"
Public Const DDE_COMMAND3 = "NORMAL"

' For some reason when performing an advise the data seems to be truncated by three
' characters. So to alleviate that problem we use a magic number constant.
Public Const MAGIC_NUMBER = 3

' This is just a string that we'll return whenever a client performs a DDE
' request if the input window on the form is empty. Otherwise we'll return
' whatever is in the window.
Public Const DDE_REQUEST_STRING = "The server input window is empty."

'*************************************************************************
' DDEML Return Values
'*************************************************************************
Public Const DMLERR_NO_ERROR = 0
Public Const DMLERR_ADVACKTIMEOUT = &H4000
Public Const DMLERR_BUSY = &H4001
Public Const DMLERR_DATAACKTIMEOUT = &H4002
Public Const DMLERR_DLL_NOT_INITIALIZED = &H4003
Public Const DMLERR_DLL_USAGE = &H4004
Public Const DMLERR_EXECACKTIMEOUT = &H4005
Public Const DMLERR_INVALIDPARAMETER = &H4006
Public Const DMLERR_LOW_MEMORY = &H4007
Public Const DMLERR_MEMORY_ERROR = &H4008
Public Const DMLERR_NOTPROCESSED = &H4009
Public Const DMLERR_NO_CONV_ESTABLISHED = &H400A
Public Const DMLERR_POKEACKTIMEOUT = &H400B
Public Const DMLERR_POSTMSG_FAILED = &H400C
Public Const DMLERR_REENTRANCY = &H400D
Public Const DMLERR_SERVER_DIED = &H400E
Public Const DMLERR_SYS_ERROR = &H400F
Public Const DMLERR_UNADVACKTIMEOUT = &H4010
Public Const DMLERR_UNFOUND_QUEUE_ID = &H4011

'*************************************************************************
' DDEML Flags
'*************************************************************************
Public Const APPCMD_FILTERINITS = &H20&
Public Const XCLASS_BOOL = &H1000
Public Const XCLASS_DATA = &H2000
Public Const XCLASS_FLAGS = &H4000
Public Const XCLASS_NOTIFICATION = &H8000
Public Const XTYPF_NOBLOCK = &H2    ' CBR_BLOCK doesn't seem to work
Public Const XTYP_ADVDATA = (&H10 Or XCLASS_FLAGS)
Public Const XTYP_ADVREQ = (&H20 Or XCLASS_DATA Or XTYPF_NOBLOCK)
Public Const XTYP_ADVSTART = (XCLASS_BOOL Or &H30)
Public Const XTYP_ADVSTOP = (XCLASS_NOTIFICATION Or &H40)
Public Const XTYP_CONNECT = (XCLASS_BOOL Or &H60 Or XTYPF_NOBLOCK)
Public Const XTYP_CONNECT_CONFIRM = (XCLASS_NOTIFICATION Or &H70 Or XTYPF_NOBLOCK)
Public Const XTYP_DISCONNECT = (XCLASS_NOTIFICATION Or &HC0 Or XTYPF_NOBLOCK)
Public Const XTYP_ERROR = (XCLASS_NOTIFICATION Or &H0 Or XTYPF_NOBLOCK)
Public Const XTYP_EXECUTE = (XCLASS_FLAGS Or &H50)
Public Const XTYP_MASK = &HF0
Public Const XTYP_MONITOR = (XCLASS_NOTIFICATION Or &HF0 Or XTYPF_NOBLOCK)
Public Const XTYP_POKE = (XCLASS_FLAGS Or &H90)
Public Const XTYP_REGISTER = (XCLASS_NOTIFICATION Or &HA0 Or XTYPF_NOBLOCK)
Public Const XTYP_REQUEST = (XCLASS_DATA Or &HB0)
Public Const XTYP_SHIFT = 4  '  shift to turn XTYP_ into an index
Public Const XTYP_UNREGISTER = (XCLASS_NOTIFICATION Or &HD0 Or XTYPF_NOBLOCK)
Public Const XTYP_WILDCONNECT = (XCLASS_DATA Or &HE0 Or XTYPF_NOBLOCK)
Public Const XTYP_XACT_COMPLETE = (XCLASS_NOTIFICATION Or &H80)
Public Const CP_WINANSI = 1004      ' Default codepage for DDE conversations.
Public Const CP_WINUNICODE = 1200
Public Const CF_TEXT = 1
Public Const CBF_SKIP_ALLNOTIFICATIONS = &H3C0000
Public Const TIMEOUT_ASYNC = &HFFFF
Public Const DNS_REGISTER = &H1
Public Const DNS_UNREGISTER = &H2
Public Const DDE_FACK = &H8000
Public Const DDE_FBUSY = &H4000
Public Const DDE_FNOTPROCESSED = &H0
Public Const HDATA_APPOWNED = &H1

'*************************************************************************
' DDEML Type Declarations
'*************************************************************************
Public Type SECURITY_QUALITY_OF_SERVICE
    Length As Long
    Impersonationlevel As Integer
    ContextTrackingMode As Integer
    EffectiveOnly As Long
End Type

Public Type CONVCONTEXT
    cb As Long
    wFlags As Long
    wCountryID As Long
    iCodePage As Long
    dwLangID As Long
    dwSecurity As Long
    qos As SECURITY_QUALITY_OF_SERVICE
End Type

'*************************************************************************
' DDEML Function Declarations
'*************************************************************************
Public Declare Function DdeInitialize Lib "user32" Alias "DdeInitializeA" _
    (pidInst As Long, _
    ByVal pfnCallback As Long, _
    ByVal afCmd As Long, _
    ByVal ulRes As Long) As Integer
    
' Removed the alias.
Public Declare Function DdeUninitialize Lib "user32" _
    (ByVal idInst As Long) As Long
    
Public Declare Function DdeNameService Lib "user32" _
            (ByVal idInst As Long, _
            ByVal hsz1 As Long, _
            ByVal hsz2 As Long, _
            ByVal afCmd As Long) As Long

' Removed the alias.
Public Declare Function DdeConnect Lib "user32" _
    (ByVal idInst As Long, _
    ByVal hszService As Long, _
    ByVal hszTopic As Long, _
    pCC As Any) As Long
    
' Removed the alias.
Public Declare Function DdeDisconnect Lib "user32" _
    (ByVal hConv As Long) As Long
    
Public Declare Function DdeCreateStringHandle Lib "user32" _
    Alias "DdeCreateStringHandleA" _
    (ByVal idInst As Long, _
    ByVal psz As String, _
    ByVal iCodePage As Long) As Long
    
' Removed the alias.
Public Declare Function DdeFreeStringHandle Lib "user32" _
    (ByVal idInst As Long, _
    ByVal hsz As Long) As Long

' Removed the alias.
Public Declare Function DdeQueryString Lib "user32" _
    Alias "DdeQueryStringA" _
    (ByVal idInst As Long, _
    ByVal hsz As Long, _
    ByVal psz As String, _
    ByVal cchMax As Long, _
    ByVal iCodePage As Long) As Long

Public Declare Function DdeCreateDataHandle Lib "user32" _
    (ByVal idInst As Long, _
    ByVal pSrc As String, _
    ByVal cb As Long, _
    ByVal cbOff As Long, _
    ByVal hszItem As Long, _
    ByVal wFmt As Long, _
    ByVal afCmd As Long) As Long
            
' Removed the alias and changed the first parameter from "ByRef pData as Byte"
' to "ByVal pData as String".
Public Declare Function DdeClientTransaction Lib "user32" _
    (ByVal pData As String, _
    ByVal cbData As Long, _
    ByVal hConv As Long, _
    ByVal hszItem As Long, _
    ByVal wFmt As Long, _
    ByVal wType As Long, _
    ByVal dwTimeout As Long, _
    pdwResult As Long) As Long
    
' The API loader provides an alias of "DdeGetDataA" for this function.
' You need to remove it because the DLL entry point can't be found for
' the alias.
Public Declare Function DdeGetData Lib "user32" _
    (ByVal hData As Long, _
    ByVal pDst As String, _
    ByVal cbMax As Long, _
    ByVal cbOff As Long) As Long

' Removed the alias.
Public Declare Function DdeFreeDataHandle Lib "user32" _
    (ByVal hData As Long) As Long

' Removed the alias.
Public Declare Function DdeGetLastError Lib "user32" _
    (ByVal idInst As Long) As Long
    
Public Declare Function DdePostAdvise Lib "user32" _
    (ByVal idInst As Long, _
    ByVal hszTopic As Long, _
    ByVal hszItem As Long) As Long
Function DDECallback(ByVal wType As Long, _
                     ByVal wFmt As Long, _
                     ByVal hConv As Long, _
                     ByVal hszTopic As Long, _
                     ByVal hszItem As Long, _
                     ByVal hData As Long, _
                     ByVal lData1 As Long, _
                     ByVal lData2 As Long) As Long
        
'    static HANDLE hToken;
'    static TOKEN_PRIVILEGES tp;
'    static LUID luid;
'    CSGTShellFunc* psh = CSGTShellFunc::Instance();
'
'
'
'    if (wType == XTYP_WILDCONNECT)
'    {
'        HDDEDATA hData;
'        HSZPAIR FAR *phszp;
'        DWORD cb;
'
'        if ((!hszTopic || hszTopic == g_hszProgman) && (!hszItem || hszItem == g_hszProgman))
'        {
'            if ((hData = DdeCreateDataHandle(g_DDEInst, NULL,
'                                             2 * sizeof(HSZPAIR), 0L, 0, 0, 0)))
'            {
'                phszp = (HSZPAIR FAR *)DdeAccessData(hData, &cb);
'                phszp[0].hszSvc = g_hszProgman;
'                phszp[0].hszTopic = g_hszProgman;
'                phszp[1].hszSvc = phszp[1].hszTopic = 0;
'                DdeUnaccessData(hData);
'                return (hData);
'            }
'        }
'        return (0);
'    }
'
    
    Dim lSize As Long
    Dim lRet As Long
    Dim sBuffer As String
    Dim sRequest As String
    
    Select Case wType
        ' Process the connect transaction.
        Case XTYP_CONNECT
            ' Just return a positive acknowledgement. If we don't the conversation will
            ' never be completed between us and the client.
            lRet = DDE_FACK
            
        Case XTYP_ADVSTART, XTYP_ADVSTOP
            If wFmt = CF_TEXT Then
                lRet = DDE_FACK
            End If
            
        ' Process the request transaction.
        Case XTYP_REQUEST, XTYP_ADVREQ

            If hszTopic = g_hszProgman And hszItem = g_hszGroups And wFmt = CF_TEXT Then
                oServer.Execute "!Log DDEServer XTYP_REQUEST or XTYP_ADVREQ"
            
                ' What's the size of the string?
                'lSize = DdeQueryString(lInstID, hszItem, vbNullString, 0, CP_WINANSI)
                
                ' Allocate space for the string.
                'sBuffer = Space(lSize)
                
                ' Grab the string.
                'DdeQueryString lInstID, hszItem, sBuffer, lSize + 1, CP_WINANSI
                            
'                LPVOID pList = NULL;
'                UINT ulLen = 0;
'
'                HDDEDATA hDDERet = (HDDEDATA) 0;
'
'                if (psh->ListGroups(pList, ulLen))
'                {
'                    hDDERet = DdeCreateDataHandle(g_DDEInst, (LPBYTE)pList, ulLen, 0L,
'                                                  g_hszGroups, CF_TEXT, 0);
'                    HeapFree(GetProcessHeap(), 0, pList);
'                }
'                return (hDDERet);
                
                
                ' Check to see if the client is requesting something we can supply.
                'If (sBuffer = DDE_REQUEST) Then
                '    If (frmDDEServer.txtData.Text = "") Then
                '        sRequest = DDE_REQUEST_STRING
                '    Else
                '        sRequest = frmDDEServer.txtData.Text
                '    End If
                    
                    ' Create a data object and populate it with our string. Then return the object.
                '    lRet = DdeCreateDataHandle(lInstID, sRequest, Len(sRequest), 0, hszItem, CF_TEXT, 0)
                'Else
                    ' The client didn't ask nicely so we're not going to process the request.
                    lRet = DDE_FNOTPROCESSED
                'End If
            End If
            
        ' Process the execute transaction.
        Case XTYP_EXECUTE
        
            If hszTopic = g_hszGroups Or hszTopic = g_hszAppProperties Then
        
                ' What's the size of the string?
                lSize = DdeGetData(ByVal hData, vbNullString, 0, 0)
                
                ' Allocate space for the buffer.
                sBuffer = Space(lSize)
                
                ' Grab the DDE data object.
                DdeGetData ByVal hData, sBuffer, lSize, 0
                
                sBuffer = UCase(sBuffer)
                
                ' Set the default return.
                lRet = DDE_FACK
                
                ' Did the client specify a command that we understand?
                ' Yes
                oServer.Execute "!log DDEServer XTYP_EXECUTE Command=" & sBuffer
                ' No
                'lRet = DDE_FNOTPROCESSED
            End If
            
    End Select
    
    ' Set the final callback return.
    DDECallback = lRet
    
End Function
Public Sub DDEInitiate()
    
    ' Initialize the DDE subsystem. We need to let the DDEML know what callback
    ' function we intend to use so we pass it address using the AddressOf operator.
    ' If we can't initialize the DDEML subsystem we exit.
    If DdeInitialize(lInstID, AddressOf DDECallback, APPCMD_FILTERINITS, 0) Then
        Exit Sub
    End If
   
    ' Now that the DDEML subsystem is initialized we create string handles for our
    ' server/topic name.
    
    g_hszProgman = DdeCreateStringHandle(lInstID, "PROGMAN", CP_WINANSI)
    g_hszGroups = DdeCreateStringHandle(lInstID, "Groups", CP_WINANSI)
    g_hszFolders = DdeCreateStringHandle(lInstID, "Folders", CP_WINANSI)
    g_hszAppProperties = DdeCreateStringHandle(lInstID, "AppProperties", CP_WINANSI)
    
    If (0 = g_hszProgman Or 0 = g_hszGroups Or 0 = g_hszFolders Or 0 = g_hszAppProperties) Then
        oServer.Execute "!Log DDEServer - Unable to create DDE string handles"
        Exit Sub
    End If
    
    ' Lets check to see if another DDE server has already registered with identical
    ' server/topic names. If so we'll exit. If we were to continue the DDE subsystem
    ' could become unstable when a client tried to converse with the server/topic.
    If (DdeConnect(lInstID, g_hszProgman, g_hszGroups, ByVal 0&)) Then
        oServer.Execute "!Log DDEServer -  A DDE server named " & Chr(34) & DDE_SERVER & Chr(34) & " with topic " & _
                Chr(34) & DDE_TOPIC & Chr(34) & " is already running."
        Exit Sub
    End If
        
    ' We need to register the server with the DDE subsystem.
    If (DdeNameService(lInstID, g_hszProgman, 0, DNS_REGISTER)) Then
        ' Set the server running flag.
        bRunning = True
    Else
        oServer.Execute "!Log DDEServer - Unable to initilise DDE PROGMAN service"
        Exit Sub
    End If
    
    If (DdeNameService(lInstID, g_hszFolders, 0, DNS_REGISTER)) Then
        ' Set the server running flag.
        bRunning = True
    Else
        oServer.Execute "!Log DDEServer - Unable to initilise DDE Folders service"
        Exit Sub

    End If
    
'    Me.WindowState = vbMinimized
End Sub
Public Sub DDETerminate()
    ' We need to release our string handles.
    DdeFreeStringHandle lInstID, g_hszProgman
    DdeFreeStringHandle lInstID, g_hszGroups
    DdeFreeStringHandle lInstID, g_hszFolders
    DdeFreeStringHandle lInstID, g_hszAppProperties
    
    ' Unregister the DDE server.
    If bRunning Then
        DdeNameService lInstID, g_hszProgman, 0, DNS_UNREGISTER
        DdeNameService lInstID, g_hszFolders, 0, DNS_UNREGISTER
    End If
    
    ' Break down the link with the DDE subsystem.
    DdeUninitialize lInstID
End Sub
